# coding: utf-8
"""
Tools to interact with Microsoft Visual C++ (MSVC) toolchain
"""

from __future__ import absolute_import as _abs
import os
import glob
import re
from pkg_resources import parse_version
from .util import _create_shared_base, _libext

LIBEXT = _libext()


def _is_64bit_windows():
    return 'PROGRAMFILES(X86)' in os.environ


def _varsall_bat_path():  # pylint: disable=R0912
    # if a custom location is given, try that first
    if 'TREELITE_VCVARSALL' in os.environ:
        candidate = os.environ['TREELITE_VCVARSALL']
        if os.path.basename(candidate).lower() != 'vcvarsall.bat':
            raise OSError('Environment variable TREELITE_VCVARSALL must point to ' + \
                          'file vcvarsall.bat')
        if os.path.isfile(candidate):
            return candidate
        raise OSError('Environment variable TREELITE_VCVARSALL does not refer ' + \
                      'to existing vcvarsall.bat')

    ## Bunch of heuristics to locate vcvarsall.bat
    candidate_paths = []  # List of possible paths to vcvarsall.bat
    try:
        import winreg  # pylint: disable=E0401
        if _is_64bit_windows():
            key_name = 'SOFTWARE\\Wow6432Node\\Microsoft\\VisualStudio\\SxS\\VS7'
        else:
            key_name = 'SOFTWARE\\Microsoft\\VisualStudio\\SxS\\VC7'
        key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, key_name)
        i = 0
        while True:
            try:
                version, vcroot, _ = winreg.EnumValue(key, i)
                if parse_version(version) >= parse_version('15.0'):
                    # Visual Studio 2017 revamped directory structure
                    candidate_paths.append(
                        os.path.join(vcroot, 'VC\\Auxiliary\\Build\\vcvarsall.bat'))
                else:
                    candidate_paths.append(os.path.join(vcroot, 'VC\\vcvarsall.bat'))
            except WindowsError:  # pylint: disable=E0602
                break
            i += 1
    except FileNotFoundError:
        pass  # No registry key found
    except ImportError:
        pass  # No winreg module

    for candidate in candidate_paths:
        if os.path.isfile(candidate):
            return candidate

    # If registry method fails, try a bunch of pre-defined paths

    # Visual Studio 2017 and higher
    for vcroot in glob.glob('C:\\Program Files (x86)\\Microsoft Visual Studio\\*') + \
                  glob.glob('C:\\Program Files\\Microsoft Visual Studio\\*'):
        if re.fullmatch(r'[0-9]+', os.path.basename(vcroot)):
            for candidate in glob.glob(vcroot + '\\*\\VC\\Auxiliary\\Build\\vcvarsall.bat'):
                if os.path.isfile(candidate):
                    return candidate
    # Previous versions of Visual Studio
    pattern = '\\Microsoft Visual Studio*\\VC\\vcvarsall.bat'
    for candidate in glob.glob('C:\\Program Files (x86)' + pattern) + \
                     glob.glob('C:\\Program Files' + pattern):
        if os.path.isfile(candidate):
            return candidate

    raise OSError('vcvarsall.bat not found; please specify its full path in ' + \
                  'the environment variable TREELITE_VCVARSALL')


def _obj_ext():
    return '.obj'


# pylint: disable=W0613
def _obj_cmd(source, toolchain, options):
    return 'cl.exe /c /openmp /Ox {} {}' \
        .format(source + '.c', ' '.join(options))


# pylint: disable=W0613
def _lib_cmd(objects, target, lib_ext, toolchain, options):
    return 'cl.exe /LD /Fe{} /openmp {} {}' \
        .format(target, ' '.join(objects), ' '.join(options))


# pylint: disable=R0913
def _create_shared(dirpath, toolchain, recipe, nthread, options, verbose):
    # Specify command to compile an object file
    recipe['object_ext'] = _obj_ext()
    recipe['library_ext'] = LIBEXT

    # pylint: disable=C0111
    def obj_cmd(source):
        return _obj_cmd(source, toolchain, options)

    def lib_cmd(objects, target):
        return _lib_cmd(objects, target, LIBEXT, toolchain, options)

    recipe['create_object_cmd'] = obj_cmd
    recipe['create_library_cmd'] = lib_cmd
    recipe['initial_cmd'] = '\"{}\" {}\n' \
        .format(_varsall_bat_path(),
                'amd64' if _is_64bit_windows() else 'x86')
    return _create_shared_base(dirpath, recipe, nthread, verbose)


__all__ = []



# coding: utf-8
"""Converter for sklearn.ensemble.RandomForestClassifier (binary classifier)"""
import treelite


class SKLRFClassifierMixin:
    """Mixin class to implement the converter for RandomForestClassifier (binary classifier)"""

    @classmethod
    def process_model(cls, sklearn_model):
        """Process a RandomForestClassifier (binary classifier) to convert it into a
           Treelite model"""
        builder = treelite.ModelBuilder(
            num_feature=sklearn_model.n_features_, average_tree_output=True,
            threshold_type='float64', leaf_output_type='float64')
        for i in range(sklearn_model.n_estimators):
            # Process i-th tree and add to the builder
            builder.append(cls.process_tree(sklearn_model.estimators_[i].tree_,
                                            sklearn_model))

        return builder.commit()

    @classmethod
    def process_leaf_node(cls, treelite_tree, sklearn_tree, node_id, sklearn_model):
        # pylint: disable=W0613
        """Process a test node with a given node ID"""
        # Get counts for each label (+/-) at this leaf node
        leaf_count = sklearn_tree.value[node_id].squeeze()
        # Compute the fraction of positive data points at this leaf node
        fraction_positive = float(leaf_count[1]) / leaf_count.sum()
        # The fraction above is now the leaf output
        treelite_tree[node_id].set_leaf_node(fraction_positive, leaf_value_type='float64')



# coding: utf-8
"""
CLI script to convert a model file to a Treelite binary checkpoint
"""
import argparse
import os
import pickle
import treelite


def main():
    """Convert a model file to a Treelite binary checkpoint"""
    parser = argparse.ArgumentParser(description='CLI script to convert a model file to a ' +
                                                 'Treelite binary checkpoint')
    parser.add_argument('--input-model', type=str, required=True,
                        help='Path to the tree model file')
    parser.add_argument('--input-model-type', type=str, required=True,
                        choices=['sklearn_pkl', 'xgboost', 'xgboost_json', 'lightgbm'],
                        help='Type of the tree model file')
    parser.add_argument('--output-checkpoint', type=str, required=True,
                        help='Path to the checkpoint file, using a special binary format for ' +
                             'Treelite')
    args = parser.parse_args()

    if args.input_model_type == 'sklearn_pkl':
        with open(args.input_model, 'rb') as f:
            sklearn_model = pickle.load(f)
        model = treelite.sklearn.import_model(sklearn_model)
    elif args.input_model_type in ['xgboost', 'xgboost_json', 'lightgbm']:
        model = treelite.Model.load(args.input_model, model_format=args.input_model_type)
    else:
        raise ValueError(f'Unrecognized type of model type: {args.input_model_type}')
    dest_checkpoint = os.fspath(os.path.expanduser(args.output_checkpoint))
    model.serialize(dest_checkpoint)


if __name__ == '__main__':
    main()



"""
General Tree Inference Library (GTIL)
"""
import ctypes
import numpy as np
from ..frontend import Model
from ..core import _LIB, _check_call


def predict(model: Model, data: np.ndarray, nthread: int = -1, pred_margin: bool = False):
    """
    Predict with a Treelite model using General Tree Inference Library (GTIL). GTIL is intended to
    be a reference implementation. GTIL is also useful in situations where using a C compiler is
    not feasible.

    .. note:: GTIL is currently experimental

        GTIL is currently in its early stage of development and may have bugs and performance
        issues. Please report any issues found on GitHub.

    Parameters
    ----------
    model : :py:class:`Model` object
        Treelite model object
    data : :py:class:`numpy.ndarray` array
        2D NumPy array, with which to run prediction
    nthread : :py:class:`int <python:int>`, optional
        Number of CPU cores to use in prediction. If <= 0, use all CPU cores.
    pred_margin : :py:class:`bool <python:bool>`, optional
        Whether to produce raw margin scores

    Returns
    -------
    prediction : :py:class:`numpy.ndarray` array
        Prediction
    """
    assert isinstance(model, Model)
    assert isinstance(data, np.ndarray)
    assert len(data.shape) == 2
    data = np.array(data, copy=False, dtype=np.float32, order='C')
    output_size = ctypes.c_size_t()
    _check_call(_LIB.TreeliteGTILGetPredictOutputSize(model.handle, ctypes.c_size_t(data.shape[0]),
                                                      ctypes.byref(output_size)))
    out_result = np.zeros(shape=output_size.value, dtype=np.float32, order='C')
    out_result_size = ctypes.c_size_t()
    _check_call(_LIB.TreeliteGTILPredict(
        model.handle,
        data.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
        ctypes.c_size_t(data.shape[0]),
        out_result.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
        ctypes.c_int(nthread),
        ctypes.c_int(0 if pred_margin else 1),
        ctypes.byref(out_result_size)
    ))
    idx = int(out_result_size.value)
    res = out_result[0:idx].reshape((data.shape[0], -1)).squeeze()
    if model.num_class > 1 and data.shape[0] != idx:
        res = res.reshape((-1, model.num_class))
    return res
